// Netify Agent Lua Processor
// Copyright (C) 2024 eGloo Incorporated
// <http://www.egloo.ca>
//
// This program is free software: you can redistribute it
// and/or modify it under the terms of the GNU General
// Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// This program is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A
// PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program.  If not, see
// <http://www.gnu.org/licenses/>.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>

#include <chrono>

#include <nd-except.hpp>
#include <nd-plugin.hpp>
#include <nd-protos.hpp>
#include <nd-sha1.h>
#include <nd-util.hpp>

#include "npp-macros.hpp"
#include "npp-plugin.hpp"

using namespace std;
using json = nlohmann::json;

#define _NPP_LOADER_CHUNK_SIZE 4096

static const char *npp_this_key = "nppPlugin::this";

static void *
lua_alloc(void *ud, void *ptr, size_t osize, size_t nsize) {
    if (nsize == 0) {
        free(ptr);
        return nullptr;
    }
    return realloc(ptr, nsize);
}

static const char *
lua_reader(lua_State *lua, void *ud, size_t *size) {
    nppPlugin *plugin = static_cast<nppPlugin *>(ud);
    return plugin->LoadNextScriptChunk(size);
}

static nppPlugin *lua_getplugin(lua_State *lua) {
    lua_pushlightuserdata(lua, static_cast<void *>(&npp_this_key));
    if (lua_gettable(lua, LUA_REGISTRYINDEX) == LUA_TNIL) {
        lua_pushfstring(lua, "plugin pointer not found: %s",
          npp_this_key);
        lua_error(lua);
    }

    nppPlugin *plugin = static_cast<nppPlugin *>(
      lua_touserdata(lua, -1));
    lua_pop(lua, 1);

    return plugin;
}

// Exported Lua function: debug(string text)
static int lua_debug(lua_State *lua) {
    nppPlugin *plugin = lua_getplugin(lua);
    const char *text = luaL_checkstring(lua, 1);

    nd_dprintf("%s: %s\n", plugin->GetTag().c_str(), text);

    return 0;
}

// Exported Lua function: print(string text)
static int lua_print(lua_State *lua) {
    nppPlugin *plugin = lua_getplugin(lua);
    const char *text = luaL_checkstring(lua, 1);

    nd_printf("%s: %s\n", plugin->GetTag().c_str(), text);

    return 0;
}

// Exported Lua function:
// dispatch(
//  string target, table channels,
//  string payload, integer format, integer compressor)
static int lua_dispatch(lua_State *lua) {
    nppPlugin *plugin = lua_getplugin(lua);

    luaL_checktype(lua, 1, LUA_TSTRING);
    luaL_checktype(lua, 2, LUA_TTABLE);
    luaL_checktype(lua, 3, LUA_TSTRING);
    luaL_checktype(lua, 4, LUA_TNUMBER);
    luaL_checktype(lua, 5, LUA_TNUMBER);

    const string target(luaL_checkstring(lua, 1));

    lua_pushnil(lua);
    ndPlugin::Channels channels;
    while (lua_next(lua, 2)) {
        channels.insert(lua_tostring(lua, -1));
        lua_pop(lua, 1);
    }

    const string payload(luaL_checkstring(lua, 3));
    const vector<uint8_t> data(payload.begin(), payload.end());

    const unsigned format = luaL_checkinteger(lua, 4);
    const unsigned compressor = luaL_checkinteger(lua, 5);

    ndFlags<ndPlugin::DispatchFlags> flags =
      ndPlugin::DispatchFlags::NONE;

    switch (format) {
    case nppPlugin::FORMAT_JSON:
        flags |= ndPlugin::DispatchFlags::FORMAT_JSON;
        break;
    case nppPlugin::FORMAT_MSGPACK:
        flags |= ndPlugin::DispatchFlags::FORMAT_MSGPACK;
        break;
    default: break;
    }

    switch (compressor) {
    case nppPlugin::COMPRESSOR_GZ:
        flags |= ndPlugin::DispatchFlags::GZ_DEFLATE;
        break;
    default: break;
    }

    try {
        plugin->DispatchSinkPayload(target, channels,
          data.size(), &data[0], flags);
    }
    catch (exception &e) {
        lua_pushfstring(lua, "dispatch exception: %s", e.what());
        lua_error(lua);
    }

    return 0;
}

nppPlugin::nppPlugin(const string &tag, const ndPlugin::Params &params)
  : ndPluginProcessor(tag, params), lua(nullptr),
    ifs_chunk(_NPP_LOADER_CHUNK_SIZE, '\0') {

    if (conf_filename.empty()) {
        throw ndExceptionSystemErrno(tag.c_str(),
          "conf_filename", EINVAL);
    }

    lua = lua_newstate(lua_alloc, static_cast<void *>(this));
    if (lua == nullptr) {
        throw ndExceptionSystemErrno(tag.c_str(),
          "lua_newstate", ENOMEM);
    }

    lua_Number lv = 0;
#if LUA_VERSION_NUM < 504
    const lua_Number *lvp = lua_version(lua);
    lv = *lvp;
#else
    lv = lua_version(lua);
#endif
    nd_dprintf("%s: initialized with Lua core v%.02f\n",
      tag.c_str(), lv / 100.0f);

    reload = true;
}

nppPlugin::~nppPlugin() {
    Join();

    if (lua != nullptr) {
        lua_close(lua);
        lua = nullptr;
    }

    nd_dprintf("%s: destroyed\n", tag.c_str());
}

void nppPlugin::GetVersion(string &version) {
    version = PACKAGE_VERSION;
}

void *nppPlugin::Entry(void) {
    nd_printf(
      "%s: %s v%s Copyright (C) 2024 eGloo Incorporated.\n",
      tag.c_str(), PACKAGE_NAME, PACKAGE_VERSION);

    while (! ShouldTerminate()) {
        if (reload.load()) {
            Reload();
            reload = false;
        }

        unique_lock<mutex> ul(event_queue_mutex);
        if (! event_queue_cond.wait_for(ul, chrono::seconds(1),
              [this] { return (! event_queue.empty()); }))
        {
            continue;
        }

        nppEvent event = event_queue.front();
        event_queue.pop();

        ul.unlock();

        switch (event.event) {
        case ndPluginProcessor::Event::UPDATE_INIT:
        case ndPluginProcessor::Event::UPDATE_COMPLETE:
            if (script_functions.load() & sfPROCESS_AGENT_EVENT)
                ProcessAgentEvent(event);
            break;
        case ndPluginProcessor::Event::FLOW_NEW:
        case ndPluginProcessor::Event::FLOW_EXPIRE:
        case ndPluginProcessor::Event::FLOW_EXPIRING:
        case ndPluginProcessor::Event::DPI_NEW:
        case ndPluginProcessor::Event::DPI_UPDATE:
        case ndPluginProcessor::Event::DPI_COMPLETE:
            if (script_functions.load() & sfPROCESS_FLOW_EVENT)
                ProcessFlowEvent(event);
            break;
        case ndPluginProcessor::Event::FLOW_MAP:
            if (script_functions.load() & sfPROCESS_FLOW_MAP)
                ProcessFlowMap(event);
            if (script_functions.load() & sfPROCESS_FLOW_STATS)
                ProcessFlowStats(event);
            break;
        default: break;
        }
    }

    return nullptr;
}

void nppPlugin::DispatchEvent(ndPlugin::Event event, void *param) {
    switch (event) {
    case ndPlugin::Event::RELOAD: reload = true; break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(
  ndPluginProcessor::Event event, ndFlowMap *flow_map) {
    if (! (script_functions.load() & sfPROCESS_FLOW_MAP) &&
      ! (script_functions.load() & sfPROCESS_FLOW_STATS))
        return;

    nppEvent e(event);
    size_t buckets = flow_map->GetBuckets();

    for (size_t b = 0; b < buckets; b++) {
        auto &fm = flow_map->Acquire(b);

        for (auto &it : fm) {
            if (! it.second->flags.detection_init.load())
                continue;
            if (it.second->flags.expired.load()) continue;

            e.PushFlow(it.second);
        }

        flow_map->Release(b);
    }

    if (e.flows.empty()) return;

    lock_guard<mutex> lg(event_queue_mutex);
    event_queue.push(e);
    event_queue_cond.notify_one();
}

void nppPlugin::DispatchProcessorEvent(
  ndPluginProcessor::Event event, nd_flow_ptr &flow) {
    if (! (script_functions.load() & sfPROCESS_FLOW_EVENT))
        return;

    lock_guard<mutex> lg(event_queue_mutex);
    event_queue.push(nppEvent(event, flow));
    event_queue_cond.notify_one();
}

void nppPlugin::DispatchProcessorEvent(
  ndPluginProcessor::Event event, ndInterfaces *interfaces) {
    // All capture interfaces and their stats are made
    // available on each update interval.
    switch (event) {
    case ndPluginProcessor::Event::INTERFACES: break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(ndPluginProcessor::Event event,
  const string &iface, ndPacketStats *stats) {
    // Capture interfaces packet stats events are dispatched
    // on each update interval.
    switch (event) {
    case ndPluginProcessor::Event::PKT_CAPTURE_STATS: break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(
  ndPluginProcessor::Event event, ndPacketStats *stats) {
    // Global (summary) capture packet stats are made
    // available on each update interval.
    switch (event) {
    case ndPluginProcessor::Event::PKT_GLOBAL_STATS: break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(
  ndPluginProcessor::Event event, ndInstanceStatus *status) {
    if (! (script_functions.load() & sfPROCESS_AGENT_EVENT))
        return;

    lock_guard<mutex> lg(event_queue_mutex);
    event_queue.push(nppEvent(event));
    event_queue_cond.notify_one();
}

void nppPlugin::DispatchProcessorEvent(ndPluginProcessor::Event event) {
    if (! (script_functions.load() & sfPROCESS_AGENT_EVENT))
        return;

    lock_guard<mutex> lg(event_queue_mutex);
    event_queue.push(nppEvent(event));
    event_queue_cond.notify_one();
}

const char *nppPlugin::LoadNextScriptChunk(size_t *length) {
    if (ifs_script.is_open()) {
        if (ifs_script.eof() || ifs_script.bad()) {
            ifs_script.close();
            *length = 0;
            return nullptr;
        }
    }
    else {
        ifs_script.open(script_filename);
        if (! ifs_script.is_open()) {
            throw ndException(
              "error opening script: %s: %s",
              script_filename.c_str(),
              strerror(ENOENT));
        }
    }

    ifs_script.read(reinterpret_cast<char *>(&ifs_chunk[0]),
      _NPP_LOADER_CHUNK_SIZE);
    *length = ifs_script.gcount();

    return reinterpret_cast<const char *>(&ifs_chunk[0]);
}

void nppPlugin::DispatchSinkPayload(const string &target,
  const ndPlugin::Channels &channels, size_t length,
  const uint8_t *payload, ndFlags<ndPlugin::DispatchFlags> flags) {
    ndPluginProcessor::DispatchSinkPayload(target, channels,
      length, payload, flags);
}

bool nppPlugin::ScriptHasFunction(const string &func) {
    lua_Debug ar;
    int rc = lua_getglobal(lua, func.c_str());

    if (rc != LUA_TFUNCTION || lua_getinfo(lua, ">S", &ar) == 0)
    {
        nd_dprintf("%s: %s: %s: not defined.\n", tag.c_str(),
          script_filename.c_str(), func.c_str());
    }
    else {
        nd_dprintf("%s: %s:%d: %s: defined\n", tag.c_str(),
          script_filename.c_str(), ar.linedefined, func.c_str());
        return true;
    }

    return false;
}

void nppPlugin::Reload(void) {
    nd_dprintf("%s: Loading configuration: %s\n",
      tag.c_str(), conf_filename.c_str());

    json j;
    ifstream ifs(conf_filename);
    if (! ifs.is_open()) {
        Unlock();
        throw ndException(
          "%s: error loading configuration: %s: %s",
          tag.c_str(), conf_filename.c_str(), strerror(ENOENT));
    }

    try {
        ifs >> j;
    }
    catch (json::exception &e) {
        Unlock();
        throw ndException(
          "%s: error loading configuration: %s: JSON "
          "exception: %s",
          tag.c_str(), conf_filename.c_str(), e.what());
    }
    catch (exception &e) {
        Unlock();
        throw ndException(
          "%s: error loading configuration: %s: %s",
          tag.c_str(), conf_filename.c_str(), e.what());
    }

    auto it = j.find("script");
    if (it == j.end() || it->type() != json::value_t::string)
    {
        throw ndException(
          "%s: error loading configuration: %s: script not "
          "found or wrong type.",
          tag.c_str(), conf_filename.c_str());
    }

    string filename;
    nd_expand_variables(it->get<string>(), filename, ndGC.conf_vars);

    vector<uint8_t> digest(SHA1_DIGEST_LENGTH, '\0');
    if (nd_sha1_file(filename, digest.data()) < 0) {
        throw ndException("%s: error hashing script: %s",
          tag.c_str(), filename.c_str());
    }

    if (! script_digest.empty() &&
      equal(digest.begin(), digest.end(), script_digest.begin()))
    {
        nd_dprintf("%s: %s: not reloading, no change.\n",
          tag.c_str(), filename.c_str());
        return;
    }

    script_filename = filename;
    script_digest.assign(digest.begin(), digest.end());

    lua_close(lua);
    lua = lua_newstate(lua_alloc, static_cast<void *>(this));
    if (lua == nullptr) {
        throw ndExceptionSystemErrno(tag.c_str(),
          "lua_newstate", ENOMEM);
    }

    luaL_openlibs(lua);

    lua_pushlightuserdata(lua, static_cast<void *>(&npp_this_key));
    lua_pushlightuserdata(lua, static_cast<void *>(this));

    lua_settable(lua, LUA_REGISTRYINDEX);

    lua_pushstring(lua, ndGC.path_config.c_str());
    lua_setglobal(lua, "NPP_PATH_CONFIG");
    lua_pushstring(lua, ndGC.path_shared_data.c_str());
    lua_setglobal(lua, "NPP_PATH_SHARED");
    lua_pushstring(lua, ndGC.path_state_volatile.c_str());
    lua_setglobal(lua, "NPP_PATH_VOLATILE");
    lua_pushstring(lua, ndGC.path_state_persistent.c_str());
    lua_setglobal(lua, "NPP_PATH_PERSISTENT");

    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::NONE));
    lua_setglobal(lua, "NPP_EVENT_NONE");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::FLOW_MAP));
    lua_setglobal(lua, "NPP_EVENT_FLOW_MAP");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::FLOW_NEW));
    lua_setglobal(lua, "NPP_EVENT_FLOW_NEW");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::FLOW_EXPIRING));
    lua_setglobal(lua, "NPP_EVENT_FLOW_EXPIRING");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::FLOW_EXPIRE));
    lua_setglobal(lua, "NPP_EVENT_FLOW_EXPIRED");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::DPI_NEW));
    lua_setglobal(lua, "NPP_EVENT_DPI_NEW");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::DPI_UPDATE));
    lua_setglobal(lua, "NPP_EVENT_DPI_UPDATE");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::DPI_COMPLETE));
    lua_setglobal(lua, "NPP_EVENT_DPI_COMPLETE");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::INTERFACES));
    lua_setglobal(lua, "NPP_EVENT_INTERFACES");
    lua_pushinteger(lua,
      static_cast<unsigned>(
        ndPluginProcessor::Event::PKT_CAPTURE_STATS));
    lua_setglobal(lua, "NPP_EVENT_PKT_CAPTURE_STATS");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::PKT_GLOBAL_STATS));
    lua_setglobal(lua, "NPP_EVENT_PKT_GLOBAL_STATS");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::UPDATE_INIT));
    lua_setglobal(lua, "NPP_EVENT_UPDATE_INIT");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndPluginProcessor::Event::UPDATE_COMPLETE));
    lua_setglobal(lua, "NPP_EVENT_UPDATE_COMPLETE");

    lua_pushinteger(lua, FORMAT_RAW);
    lua_setglobal(lua, "NPP_FORMAT_RAW");
    lua_pushinteger(lua, FORMAT_JSON);
    lua_setglobal(lua, "NPP_FORMAT_JSON");
    lua_pushinteger(lua, FORMAT_MSGPACK);
    lua_setglobal(lua, "NPP_FORMAT_MSGPACK");

    lua_pushinteger(lua, COMPRESSOR_NONE);
    lua_setglobal(lua, "NPP_COMPRESSOR_NONE");
    lua_pushinteger(lua, COMPRESSOR_GZ);
    lua_setglobal(lua, "NPP_COMPRESSOR_GZIP");

    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::LowerMap::UNKNOWN));
    lua_setglobal(lua, "NPP_LOWER_UNKNOWN");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::LowerMap::LOCAL));
    lua_setglobal(lua, "NPP_LOWER_LOCAL");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::LowerMap::OTHER));
    lua_setglobal(lua, "NPP_LOWER_OTHER");

    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::OtherType::UNKNOWN));
    lua_setglobal(lua, "NPP_OTHER_UNKNOWN");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::OtherType::UNSUPPORTED));
    lua_setglobal(lua, "NPP_OTHER_UNSUPPORTED");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::OtherType::LOCAL));
    lua_setglobal(lua, "NPP_OTHER_LOCAL");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::OtherType::MULTICAST));
    lua_setglobal(lua, "NPP_OTHER_MULTICAST");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::OtherType::BROADCAST));
    lua_setglobal(lua, "NPP_OTHER_BROADCAST");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::OtherType::REMOTE));
    lua_setglobal(lua, "NPP_OTHER_REMOTE");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::OtherType::ERROR));
    lua_setglobal(lua, "NPP_OTHER_ERROR");

    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::TunnelType::NONE));
    lua_setglobal(lua, "NPP_TUNNEL_NONE");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::TunnelType::GTP));
    lua_setglobal(lua, "NPP_TUNNEL_GTP");

    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::Origin::UNKNOWN));
    lua_setglobal(lua, "NPP_ORIGIN_UNKNOWN");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::Origin::LOWER));
    lua_setglobal(lua, "NPP_ORIGIN_LOWER");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndFlow::Origin::UPPER));
    lua_setglobal(lua, "NPP_ORIGIN_UPPER");

    lua_pushinteger(lua, static_cast<unsigned>(ndAddr::Type::NONE));
    lua_setglobal(lua, "NPP_ADDRTYPE_NONE");
    lua_pushinteger(lua, static_cast<unsigned>(ndAddr::Type::LOCAL));
    lua_setglobal(lua, "NPP_ADDRTYPE_LOCAL");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndAddr::Type::LOCALNET));
    lua_setglobal(lua, "NPP_ADDRTYPE_LOCALNET");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndAddr::Type::RESERVED));
    lua_setglobal(lua, "NPP_ADDRTYPE_RESERVED");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndAddr::Type::MULTICAST));
    lua_setglobal(lua, "NPP_ADDRTYPE_MULTICAST");
    lua_pushinteger(lua,
      static_cast<unsigned>(ndAddr::Type::BROADCAST));
    lua_setglobal(lua, "NPP_ADDRTYPE_BROADCAST");
    lua_pushinteger(lua, static_cast<unsigned>(ndAddr::Type::OTHER));
    lua_setglobal(lua, "NPP_ADDRTYPE_OTHER");
    lua_pushinteger(lua, static_cast<unsigned>(ndAddr::Type::ERROR));
    lua_setglobal(lua, "NPP_ADDRTYPE_ERROR");

#ifdef IPPROTO_AH
    lua_pushinteger(lua, IPPROTO_AH);
    lua_setglobal(lua, "NPP_IPPROTO_AH");
#endif
#ifdef IPPROTO_BEETPH
    lua_pushinteger(lua, IPPROTO_BEETPH);
    lua_setglobal(lua, "NPP_IPPROTO_BEETPH");
#endif
#ifdef IPPROTO_COMP
    lua_pushinteger(lua, IPPROTO_COMP);
    lua_setglobal(lua, "NPP_IPPROTO_COMP");
#endif
#ifdef IPPROTO_DCCP
    lua_pushinteger(lua, IPPROTO_DCCP);
    lua_setglobal(lua, "NPP_IPPROTO_DCCP");
#endif
#ifdef IPPROTO_EGP
    lua_pushinteger(lua, IPPROTO_EGP);
    lua_setglobal(lua, "NPP_IPPROTO_EGP");
#endif
#ifdef IPPROTO_ENCAP
    lua_pushinteger(lua, IPPROTO_ENCAP);
    lua_setglobal(lua, "NPP_IPPROTO_ENCAP");
#endif
#ifdef IPPROTO_ESP
    lua_pushinteger(lua, IPPROTO_ESP);
    lua_setglobal(lua, "NPP_IPPROTO_ESP");
#endif
#ifdef IPPROTO_ETHERNET
    lua_pushinteger(lua, IPPROTO_ETHERNET);
    lua_setglobal(lua, "NPP_IPPROTO_ETHERNET");
#endif
#ifdef IPPROTO_GRE
    lua_pushinteger(lua, IPPROTO_GRE);
    lua_setglobal(lua, "NPP_IPPROTO_GRE");
#endif
#ifdef IPPROTO_ICMP
    lua_pushinteger(lua, IPPROTO_ICMP);
    lua_setglobal(lua, "NPP_IPPROTO_ICMP");
#endif
#ifdef IPPROTO_IDP
    lua_pushinteger(lua, IPPROTO_IDP);
    lua_setglobal(lua, "NPP_IPPROTO_IDP");
#endif
#ifdef IPPROTO_IGMP
    lua_pushinteger(lua, IPPROTO_IGMP);
    lua_setglobal(lua, "NPP_IPPROTO_IGMP");
#endif
#ifdef IPPROTO_IP
    lua_pushinteger(lua, IPPROTO_IP);
    lua_setglobal(lua, "NPP_IPPROTO_IP");
#endif
#ifdef IPPROTO_IPIP
    lua_pushinteger(lua, IPPROTO_IPIP);
    lua_setglobal(lua, "NPP_IPPROTO_IPIP");
#endif
#ifdef IPPROTO_IPV6
    lua_pushinteger(lua, IPPROTO_IPV6);
    lua_setglobal(lua, "NPP_IPPROTO_IPV6");
#endif
#ifdef IPPROTO_L2TP
    lua_pushinteger(lua, IPPROTO_L2TP);
    lua_setglobal(lua, "NPP_IPPROTO_L2TP");
#endif
#ifdef IPPROTO_MPLS
    lua_pushinteger(lua, IPPROTO_MPLS);
    lua_setglobal(lua, "NPP_IPPROTO_MPLS");
#endif
#ifdef IPPROTO_MPTCP
    lua_pushinteger(lua, IPPROTO_MPTCP);
    lua_setglobal(lua, "NPP_IPPROTO_MPTCP");
#endif
#ifdef IPPROTO_MTP
    lua_pushinteger(lua, IPPROTO_MTP);
    lua_setglobal(lua, "NPP_IPPROTO_MTP");
#endif
#ifdef IPPROTO_PIM
    lua_pushinteger(lua, IPPROTO_PIM);
    lua_setglobal(lua, "NPP_IPPROTO_PIM");
#endif
#ifdef IPPROTO_PUP
    lua_pushinteger(lua, IPPROTO_PUP);
    lua_setglobal(lua, "NPP_IPPROTO_PUP");
#endif
#ifdef IPPROTO_RAW
    lua_pushinteger(lua, IPPROTO_RAW);
    lua_setglobal(lua, "NPP_IPPROTO_RAW");
#endif
#ifdef IPPROTO_RSVP
    lua_pushinteger(lua, IPPROTO_RSVP);
    lua_setglobal(lua, "NPP_IPPROTO_RSVP");
#endif
#ifdef IPPROTO_SCTP
    lua_pushinteger(lua, IPPROTO_SCTP);
    lua_setglobal(lua, "NPP_IPPROTO_SCTP");
#endif
#ifdef IPPROTO_TCP
    lua_pushinteger(lua, IPPROTO_TCP);
    lua_setglobal(lua, "NPP_IPPROTO_TCP");
#endif
#ifdef IPPROTO_TP
    lua_pushinteger(lua, IPPROTO_TP);
    lua_setglobal(lua, "NPP_IPPROTO_TP");
#endif
#ifdef IPPROTO_UDP
    lua_pushinteger(lua, IPPROTO_UDP);
    lua_setglobal(lua, "NPP_IPPROTO_UDP");
#endif
#ifdef IPPROTO_UDPLITE
    lua_pushinteger(lua, IPPROTO_UDPLITE);
    lua_setglobal(lua, "NPP_IPPROTO_UDPLITE");
#endif

    static const luaL_Reg nppFunctions[] = { { "debug", lua_debug },
        { "print", lua_print },
        { "dispatch", lua_dispatch }, { nullptr, nullptr } };

    luaL_newlib(lua, nppFunctions);
    lua_setglobal(lua, "npp");

    int rc;
    if ((rc = lua_load(lua, lua_reader,
           static_cast<void *>(this), tag.c_str(), "bt")) != LUA_OK)
    {
        auto e = lua_errors.find(rc);
        throw ndException(
          "%s: error loading script: %s: %s", tag.c_str(),
          script_filename.c_str(),
          (e == lua_errors.end()) ? "Unknown" : e->second.c_str());
    }

    if ((rc = lua_pcall(lua, 0, 0, 0)) != LUA_OK) {
        auto e = lua_errors.find(rc);
        throw ndException("%s: initialization: %s: %s",
          script_filename.c_str(),
          (e == lua_errors.end()) ? "Unknown" : e->second.c_str(),
          lua_tostring(lua, -1));
    }

    script_functions = sfNONE;
    if (ScriptHasFunction("ProcessAgentEvent"))
        script_functions |= sfPROCESS_AGENT_EVENT;
    if (ScriptHasFunction("ProcessFlowEvent"))
        script_functions |= sfPROCESS_FLOW_EVENT;
    if (ScriptHasFunction("ProcessFlowMap"))
        script_functions |= sfPROCESS_FLOW_MAP;
    if (ScriptHasFunction("ProcessFlowStats"))
        script_functions |= sfPROCESS_FLOW_STATS;

    if (script_functions == sfNONE) {
        throw ndException(
          "%s: %s: no script functions defined",
          tag.c_str(), script_filename.c_str());
    }

    lock_guard<mutex> lg(event_queue_mutex);
    while (! event_queue.empty()) event_queue.pop();
}

void nppPlugin::ProcessAgentEvent(nppEvent &event) {
    int rc = lua_getglobal(lua, __func__);

    if (rc != LUA_TFUNCTION) {
        throw ndException("%s: %s: %s function not found.",
          tag.c_str(), __func__, script_filename.c_str());
    }

    lua_pushinteger(lua, (lua_Integer)event.event);

    if ((rc = lua_pcall(lua, 1, 0, 0)) != LUA_OK) {
        auto e = lua_errors.find(rc);
        nd_dprintf("%s: %s: %s: %s\n", tag.c_str(), __func__,
          (e == lua_errors.end()) ? "Unknown" : e->second.c_str(),
          lua_tostring(lua, -1));
    }
}

void nppPlugin::PushFlowTable(const nd_flow_ptr &flow,
  const ndFlowStats &stats, bool stats_only) {
    lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

    lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

    NPP_LUA_TABLE_STRING("name", flow->iface->ifname.c_str())
    NPP_LUA_TABLE_INTEGER("role",
      static_cast<unsigned>(flow->iface->role))

    lua_setfield(lua, -2, "iface");

    string lower_digest;
    nd_sha1_to_string(flow->digest_lower, lower_digest);
    NPP_LUA_TABLE_STRING("digest_lower", lower_digest.c_str());
    string mdata_digest;
    nd_sha1_to_string(flow->digest_mdata, mdata_digest);
    NPP_LUA_TABLE_STRING("digest_mdata", mdata_digest.c_str());

    lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

    NPP_LUA_TABLE_INTEGER("lower_bytes", stats.lower_bytes.load());
    NPP_LUA_TABLE_INTEGER("upper_bytes", stats.upper_bytes.load());
    NPP_LUA_TABLE_INTEGER("total_bytes", stats.total_bytes.load());
    NPP_LUA_TABLE_INTEGER("lower_packets",
      stats.lower_packets.load());
    NPP_LUA_TABLE_INTEGER("upper_packets",
      stats.upper_packets.load());
    NPP_LUA_TABLE_INTEGER("total_packets",
      stats.total_packets.load());
    NPP_LUA_TABLE_INTEGER("detection_packets",
      stats.detection_packets.load());
#ifdef _ND_ENABLE_EXTENDED_STATS
    NPP_LUA_TABLE_NUMBER("lower_rate", stats.lower_rate.load());
    NPP_LUA_TABLE_NUMBER("upper_rate", stats.upper_rate.load());
    NPP_LUA_TABLE_INTEGER("tcp_seq_errors",
      stats.tcp_seq_errors.load());
    NPP_LUA_TABLE_INTEGER("tcp_resets", stats.tcp_resets.load());
    NPP_LUA_TABLE_INTEGER("tcp_retrans", stats.tcp_retrans.load());
#endif
    lua_setfield(lua, -2, "stats");

    if (stats_only) return;

    NPP_LUA_TABLE_INTEGER("ip_version", flow->ip_version)
    NPP_LUA_TABLE_INTEGER("ip_protocol", flow->ip_protocol)
    NPP_LUA_TABLE_INTEGER("vlan_id", flow->vlan_id)
    NPP_LUA_TABLE_INTEGER("first_seen_at", flow->ts_first_seen)
    NPP_LUA_TABLE_INTEGER("last_seen_at", flow->ts_last_seen.load())
    NPP_LUA_TABLE_INTEGER("lower_map",
      static_cast<unsigned>(flow->lower_map))
    NPP_LUA_TABLE_INTEGER("other_type",
      static_cast<unsigned>(flow->other_type))
    NPP_LUA_TABLE_INTEGER("tunnel_type",
      static_cast<unsigned>(flow->tunnel_type))
    NPP_LUA_TABLE_INTEGER("origin",
      static_cast<unsigned>(flow->origin))
    NPP_LUA_TABLE_INTEGER("privacy_mask",
      static_cast<unsigned>(flow->privacy_mask.flags))

    NPP_LUA_TABLE_STRING("lower_mac",
      flow->lower_mac.GetString().c_str())
    NPP_LUA_TABLE_STRING("upper_mac",
      flow->upper_mac.GetString().c_str())

    NPP_LUA_TABLE_STRING("lower_addr",
      flow->lower_addr.GetString().c_str())
    NPP_LUA_TABLE_STRING("upper_addr",
      flow->upper_addr.GetString().c_str())

    NPP_LUA_TABLE_INTEGER("lower_port", flow->lower_addr.GetPort())
    NPP_LUA_TABLE_INTEGER("upper_port", flow->upper_addr.GetPort())

    lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

    NPP_LUA_TABLE_BOOLEAN("detection_complete",
      flow->flags.detection_complete.load());
    NPP_LUA_TABLE_BOOLEAN("detection_guessed",
      flow->flags.detection_guessed.load());
    NPP_LUA_TABLE_BOOLEAN("detection_init",
      flow->flags.detection_init.load());
    NPP_LUA_TABLE_BOOLEAN("detection_updated",
      flow->flags.detection_updated.load());
    NPP_LUA_TABLE_BOOLEAN("dhc_hit", flow->flags.dhc_hit.load());
    NPP_LUA_TABLE_BOOLEAN("fhc_hit", flow->flags.fhc_hit.load());
    NPP_LUA_TABLE_BOOLEAN("expiring", flow->flags.expiring.load());
    NPP_LUA_TABLE_BOOLEAN("expired", flow->flags.expired.load());
    NPP_LUA_TABLE_BOOLEAN("soft_dissector",
      flow->flags.soft_dissector.load());

    lua_setfield(lua, -2, "flags");

    if (flow->ip_protocol == IPPROTO_TCP) {
        lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

        NPP_LUA_TABLE_INTEGER("fin_acks", flow->tcp.fin_ack.load())
        NPP_LUA_TABLE_INTEGER("last_seq", flow->tcp.last_seq.load())

        lua_setfield(lua, -2, "tcp");
    }

    // XXX: Lock before accessing metadata...
    flow->lock.lock();

#if defined(_ND_ENABLE_CONNTRACK) && defined(_ND_WITH_CONNTRACK_MDATA)
    lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

    NPP_LUA_TABLE_INTEGER("id", flow->conntrack.id)
    NPP_LUA_TABLE_INTEGER("mark", flow->conntrack.mark)

    lua_setfield(lua, -2, "conntrack");
#endif

    if (flow->tunnel_type == ndFlow::TunnelType::GTP) {
        lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

        NPP_LUA_TABLE_INTEGER("version", flow->gtp.version)
        NPP_LUA_TABLE_INTEGER("ip_version", flow->gtp.ip_version)

        NPP_LUA_TABLE_INTEGER("lower_teid", flow->gtp.lower_teid)
        NPP_LUA_TABLE_INTEGER("upper_teid", flow->gtp.upper_teid)

        NPP_LUA_TABLE_INTEGER("lower_type",
          static_cast<unsigned>(flow->gtp.lower_type))
        NPP_LUA_TABLE_INTEGER("upper_type",
          static_cast<unsigned>(flow->gtp.upper_type))

        NPP_LUA_TABLE_STRING("lower_addr",
          flow->gtp.lower_addr.GetString().c_str())
        NPP_LUA_TABLE_STRING("upper_addr",
          flow->gtp.upper_addr.GetString().c_str())

        NPP_LUA_TABLE_INTEGER("lower_port",
          flow->gtp.lower_addr.GetPort())
        NPP_LUA_TABLE_INTEGER("upper_port",
          flow->gtp.upper_addr.GetPort())

        NPP_LUA_TABLE_INTEGER("lower_map",
          static_cast<unsigned>(flow->gtp.lower_map))
        NPP_LUA_TABLE_INTEGER("other_type",
          static_cast<unsigned>(flow->gtp.other_type))

        lua_setfield(lua, -2, "gtp");
    }

    NPP_LUA_TABLE_STRING("dns_host_name",
      flow->dns_host_name.c_str())
    NPP_LUA_TABLE_STRING("host_server_name",
      flow->host_server_name.c_str())

    NPP_LUA_TABLE_INTEGER("detected_protocol",
      static_cast<unsigned>(flow->detected_protocol))
    NPP_LUA_TABLE_INTEGER("detected_application",
      flow->detected_application)

    NPP_LUA_TABLE_STRING("detected_protocol_name",
      flow->detected_protocol_name.c_str())
    NPP_LUA_TABLE_STRING("detected_application_name",
      flow->detected_application_name.c_str())

    lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

    NPP_LUA_TABLE_INTEGER("application", flow->category.application)
    NPP_LUA_TABLE_INTEGER("protocol", flow->category.protocol)
    NPP_LUA_TABLE_INTEGER("domain", flow->category.domain)
    NPP_LUA_TABLE_INTEGER("network", flow->category.network)

    lua_setfield(lua, -2, "category");

    switch (flow->GetMasterProtocol()) {
    case ndProto::Id::HTTP:
        lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

        NPP_LUA_TABLE_STRING("user_agent",
          flow->http.user_agent.c_str())
        NPP_LUA_TABLE_STRING("url", flow->http.url.c_str())

        lua_setfield(lua, -2, "http");
        break;

    case ndProto::Id::DHCP:
        lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

        NPP_LUA_TABLE_STRING("fingerprint",
          flow->dhcp.fingerprint.c_str())
        NPP_LUA_TABLE_STRING("class_ident",
          flow->dhcp.class_ident.c_str())

        lua_setfield(lua, -2, "dhcp");
        break;

    case ndProto::Id::SSH:
        lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

        NPP_LUA_TABLE_STRING("client_agent",
          flow->ssh.client_agent.c_str())
        NPP_LUA_TABLE_STRING("server_agent",
          flow->ssh.server_agent.c_str())

        lua_setfield(lua, -2, "ssh");
        break;

    case ndProto::Id::TLS:
        lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

        NPP_LUA_TABLE_INTEGER("version", flow->tls.version)
        NPP_LUA_TABLE_INTEGER("cipher_suite", flow->tls.cipher_suite)

        NPP_LUA_TABLE_STRING("subject_dn",
          flow->tls.subject_dn.c_str())
        NPP_LUA_TABLE_STRING("issuer_dn",
          flow->tls.issuer_dn.c_str())
        NPP_LUA_TABLE_STRING("server_cn",
          flow->tls.server_cn.c_str())
        NPP_LUA_TABLE_STRING("client_sni",
          flow->tls.client_sni.c_str())

        NPP_LUA_TABLE_STRING("client_ja3",
          flow->tls.client_ja3.c_str())
        NPP_LUA_TABLE_STRING("server_ja3",
          flow->tls.server_ja3.c_str())

        if (! flow->tls.alpn.empty()) {
            lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

            size_t i = 0;
            for (auto &it : flow->tls.alpn) {
                lua_pushinteger(lua, ++i);
                lua_pushstring(lua, it.c_str());
                lua_settable(lua, -3);
            }

            lua_setfield(lua, -2, "alpn_client");
        }

        if (! flow->tls.alpn_server.empty()) {
            lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

            size_t i = 0;
            for (auto &it : flow->tls.alpn_server) {
                lua_pushinteger(lua, ++i);
                lua_pushstring(lua, it.c_str());
                lua_settable(lua, -3);
            }

            lua_setfield(lua, -2, "alpn_server");
        }

        if (flow->HasTLSEncryptedCH()) {
            lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

            NPP_LUA_TABLE_INTEGER("version", flow->tls.ech.version)

            lua_setfield(lua, -2, "encrypted_ch");
        }

        if (flow->HasTLSEncryptedSNI()) {
            lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

            NPP_LUA_TABLE_INTEGER("cipher_suite",
              flow->tls.esni.cipher_suite)
            NPP_LUA_TABLE_STRING("client_sni",
              flow->tls.esni.esni.c_str())

            lua_setfield(lua, -2, "encrypted_sni");
        }

        lua_setfield(lua, -2, "tls");
        break;

    case ndProto::Id::MAIL_SMTP:
        lua_createtable(lua, 0 /* narr */, 0 /* nrec */);
        NPP_LUA_TABLE_BOOLEAN("tls", flow->smtp.tls)
        lua_setfield(lua, -2, "smtp");
        break;

    case ndProto::Id::MDNS:
        if (flow->HasMDNSDomainName()) {
            lua_createtable(lua, 0 /* narr */, 0 /* nrec */);
            NPP_LUA_TABLE_STRING("domain_name",
              flow->mdns.domain_name.c_str())
            lua_setfield(lua, -2, "mdns");
        }
        break;

    default: break;
    }

    if (! flow->risk.risks.empty()) {
        lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

        NPP_LUA_TABLE_INTEGER("ndpi_score", flow->risk.ndpi_score)
        NPP_LUA_TABLE_INTEGER("ndpi_score_client",
          flow->risk.ndpi_score_client)
        NPP_LUA_TABLE_INTEGER("ndpi_score_server",
          flow->risk.ndpi_score_server)

        lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

        size_t i = 0;
        for (auto &it : flow->risk.risks) {
            lua_pushinteger(lua, ++i);
            lua_pushinteger(lua, static_cast<unsigned>(it));
            lua_settable(lua, -3);
        }

        lua_setfield(lua, -2, "risks");

        lua_setfield(lua, -2, "risk");
    }

    flow->lock.unlock();
}

void nppPlugin::ProcessFlowEvent(nppEvent &event) {
    int rc = lua_getglobal(lua, __func__);

    if (rc != LUA_TFUNCTION) {
        throw ndException("%s: %s: %s function not found.",
          tag.c_str(), __func__, script_filename.c_str());
    }

    lua_pushinteger(lua, (lua_Integer)event.event);

    lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

    size_t i = 0;
    for (auto &it : event.flows) {
        lua_pushinteger(lua, ++i);
        PushFlowTable(it.first, it.second);
        lua_settable(lua, -3);
    }

    if ((rc = lua_pcall(lua, 2, 0, 0)) != LUA_OK) {
        auto e = lua_errors.find(rc);
        nd_dprintf("%s: %s: %s: %s\n", tag.c_str(), __func__,
          (e == lua_errors.end()) ? "Unknown" : e->second.c_str(),
          lua_tostring(lua, -1));
    }
}

void nppPlugin::ProcessFlowMap(nppEvent &event) {
    int rc = lua_getglobal(lua, __func__);

    if (rc != LUA_TFUNCTION) {
        throw ndException("%s: %s: %s function not found.",
          tag.c_str(), __func__, script_filename.c_str());
    }

    lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

    size_t i = 0;
    for (auto &it : event.flows) {
        lua_pushinteger(lua, ++i);
        PushFlowTable(it.first, it.second);
        lua_settable(lua, -3);
    }

    if ((rc = lua_pcall(lua, 1, 0, 0)) != LUA_OK) {
        auto e = lua_errors.find(rc);
        nd_dprintf("%s: %s: %s: %s\n", tag.c_str(), __func__,
          (e == lua_errors.end()) ? "Unknown" : e->second.c_str(),
          lua_tostring(lua, -1));
    }
}

void nppPlugin::ProcessFlowStats(nppEvent &event) {
    int rc = lua_getglobal(lua, __func__);

    if (rc != LUA_TFUNCTION) {
        throw ndException("%s: %s: %s function not found.",
          tag.c_str(), __func__, script_filename.c_str());
    }

    lua_createtable(lua, 0 /* narr */, 0 /* nrec */);

    size_t i = 0;
    for (auto &it : event.flows) {
        lua_pushinteger(lua, ++i);
        PushFlowTable(it.first, it.second, true);
        lua_settable(lua, -3);
    }

    if ((rc = lua_pcall(lua, 1, 0, 0)) != LUA_OK) {
        auto e = lua_errors.find(rc);
        nd_dprintf("%s: %s: %s: %s\n", tag.c_str(), __func__,
          (e == lua_errors.end()) ? "Unknown" : e->second.c_str(),
          lua_tostring(lua, -1));
    }
}

ndPluginInit(nppPlugin);
